/*jslint browser: true, unparam: true, todo: true, evil: true*/
/*globals define: false, performance: false, requestAnimationFrame: false, Event: false*/
define(['./components/index.js', './util.js', './dialog.js'], function (defaultComponents, util, dialog) {
    'use strict';
    var attributeList = ['data', 'schema', 'name', 'components'];
    function PivotForm(args) {
        args = args || {};
        // only non-dom related items can be instantiated prior to init (connectedCallback)
        var self = {}, intf;
        self.args = args;
        self.initialized = false;
        self.form = util.createElement('form', null, {className: 'pivot-form'});
        self.legend = util.createElement('legend', null, {className: 'legend'});
        self.uniqueId = Math.random() + '-' + performance.now();
        self.title = '';
        self.componentOrder = [];
        self.initializingComponents = {};
        // represents data input via data setter prior to async init values completed
        self.preInitData = {};
        self.components = [];
        self.idCounter = -1;
        self.data = {};
        self.dialogOptions = {};
        self.childForms = [];
        intf = window.customElements ? eval('Reflect.construct(HTMLElement, [], new.target)')
            : util.createElement('section');
        function getNewId() {
            self.idCounter += 1;
            return 'p' + Math.floor((Math.random() * Math.pow(2, 30))).toString(36);
        }
        function getComponentsByPropertyValue(key, value) {
            var found = [];
            intf.childForms.forEach(function (form) {
                found = found.concat(form.getComponentsByPropertyValue(key, value));
            });
            self.components.forEach(function (component) {
                if (component && component.isContainer) {
                    // this list of components will be in the above child forms
                    return;
                }
                if (value === component[key] || (component.header && value === component.header[key])) {
                    found.push(component);
                }
            });
            return found;
        }
        function getComponentByName(value) {
            return getComponentsByPropertyValue('name', value)[0];
        }
        function getComponentById(value) {
            return getComponentsByPropertyValue('id', value)[0];
        }
        function setPreInitData() {
            if (!self.initialized) { return; }
            if (Object.keys(self.preInitData).length > 0) {
                intf.data = self.preInitData;
            }
        }
        function saveState() {
            if (self.name) {
                try {
                    localStorage.setItem('pivot-form-order-' + self.name, JSON.stringify({
                        componentOrder: self.componentOrder
                    }));
                } catch (ignore) {
                    self.componentOrder = [];
                }
            }
        }
        function updateFlexOrder() {
            if (!self.inputComponentOrganicOrder) { return; }
            self.inputComponentOrganicOrder.forEach(function (itemContainer, organicComponentIndex) {
                organicComponentIndex += 1;
                itemContainer.style.order = self.componentOrder.indexOf(organicComponentIndex);
            });
            saveState();
        }
        function loadState() {
            if (self.name && intf.initialized) {
                try {
                    var d = localStorage.getItem('pivot-form-order-' + self.name);
                    d = self.componentOrder = JSON.parse(d).componentOrder;
                } catch (ignore) {
                    self.componentOrder = [];
                }
                updateFlexOrder();
            }
        }
        function reorderComponents(e) {
            // move up e until a base component item is found, then draw a line below it
            // and mark it as the target of the dropping (item reordering)
            self.moveMarker.style.left = e.layerX + 'px';
            self.moveMarker.style.top = e.layerY + 'px';
            var tOrd, uid, el = e.target;
            while (el && el.parentNode && el.tagName !== 'BODY') {
                tOrd = parseInt(self.moveFrom.style.order, 10);
                uid = el.getAttribute('data-uniqueId');
                if (uid === self.uniqueId && el !== self.moveFrom) {
                    if (self.moveTo) {
                        self.moveTo.classList.remove('reorder-drag-over');
                    }
                    self.moveTo = el;
                    self.moveTo.classList.add('reorder-drag-over');
                    self.moveTargetIndex = parseInt(tOrd, 10);
                    return;
                }
                el = el.parentNode;
            }
            return;
        }
        function endReorderComponents() {
            function reorder() {
                if (!self.moveTo) { return; }
                var o = self.componentOrder.indexOf(self.moveFrom.organicComponentIndex),
                    t = self.componentOrder.indexOf(self.moveTo.organicComponentIndex);
                self.componentOrder.splice(o, 1);
                self.componentOrder.splice(t, 0, self.moveFrom.organicComponentIndex);
                updateFlexOrder();
                return;
            }
            if (self.moveTo) {
                self.moveTo.classList.remove('reorder-drag-over');
            }
            reorder();
            if (self.moveMarker.parentNode) {
                self.moveMarker.parentNode.removeChild(self.moveMarker);
            }
            self.form.classList.remove('noselect');
            self.moveTo = undefined;
            self.moveFrom = undefined;
            self.moveMarker = undefined;
            self.shadowRoot.removeEventListener('mousemove', reorderComponents);
            self.shadowRoot.removeEventListener('mouseup', endReorderComponents);
            return;
        }
        function beginReorderComponent(organicComponentIndex, itemContainer) {
            return function (e) {
                self.form.classList.add('noselect');
                self.moveMarker = util.createElement('div', self.shadowRoot,
                        {className: 'reorder-marker', innerHTML: '\u205E'});
                self.moveFrom = itemContainer;
                self.moveStart = {
                    item: itemContainer.getBoundingClientRect(),
                    clientX: e.clientX,
                    clientY: e.clientY
                };
                reorderComponents(e);
                self.shadowRoot.addEventListener('mousemove', reorderComponents);
                self.shadowRoot.addEventListener('mouseup', endReorderComponents);
            };
        }
        function doneInitializing() {
            if (Object.keys(self.initializingComponents).length === 0) {
                // notify that everything has been initialized on this form
                intf.classList.remove('loading');
                intf.dispatchEvent(new Event('initialized'));
                self.initialized = true;
            }
        }
        function createDom() {
            self.components.forEach(function (component, organicComponentIndex) {
                organicComponentIndex += 1;
                var container = util.createElement('div', null, {className: 'component-item'}),
                    handle = util.createElement('span', component.header.moveable
                        ? container : null, {className: 'item-handle', innerHTML: '\u205E'}),
                    dataStub = {};
                if (component.header.hidden) {
                    container.style.display = 'none';
                }
                function initialized() {
                    component.removeEventListener('initialized', initialized);
                    delete self.initializingComponents[component.id];
                    doneInitializing();
                    setPreInitData();
                }
                component.container = container;
                component.handle = handle;
                container.setAttribute('data-uniqueId', self.uniqueId);
                handle.beginReorderComponentFunction = beginReorderComponent(organicComponentIndex, container);
                handle.addEventListener('mousedown', handle.beginReorderComponentFunction);
                container.organicComponentIndex = organicComponentIndex;
                container.name = component.name;
                self.componentOrder[organicComponentIndex] = self.componentOrder[organicComponentIndex] === undefined
                    ? organicComponentIndex : self.componentOrder[organicComponentIndex];
                container.appendChild(component);
                util.setProperties(container.style, component.containerStyle);
                util.setProperties(handle, component.handleStyle);
                util.setProperties(container.style, component.containerStyle);
                util.setProperties(handle, component.handleStyle);
                self.form.appendChild(container);
                self.inputComponentOrganicOrder.push(container);
                // a value was defined in the schema as a default
                // assume this is an async function and delay loading of data via data setter
                // by adding items to self.initializingComponents
                if (!component.initialzied && component.header.name && component.header.value) {
                    dataStub[component.header.name] = component.header.value;
                    self.initializingComponents[component.id] = component;
                    component.addEventListener('initialized', initialized);
                    component.value = dataStub;
                    return;
                }
                // if a component does not have initial data, call initialized
                // after just a moment to allow additional components in the array
                // to fill self.initializingComponents before it is checked for the first time.
                requestAnimationFrame(initialized);
            });
            updateFlexOrder();
        }
        function initData(newData) {
            function componentDataSetter(component) {
                component.value = newData;
            }
            self.components.forEach(componentDataSetter);
            intf.childForms.forEach(function (form) {
                form.components.forEach(componentDataSetter);
            });
        }
        function disposeComponents() {
            self.components.forEach(function (component) {
                if (component) {
                    if (component.dispose) {
                        component.dispose();
                    }
                    if (component.parentNode) {
                        component.parentNode.removeChild(component);
                    }
                    if (component.handle && component.handle.parentNode) {
                        component.handle.removeEventListener('mousedown', component.handle.beginReorderComponentFunction);
                        component.handle.parentNode.removeChild(component.handle);
                    }
                    if (component.container && component.container.parentNode) {
                        component.container.parentNode.removeChild(component.container);
                    }
                }
            });
        }
        function dispose() {
            disposeComponents();
        }
        function initSchema() {
            if (!self.schema) { return; }
            if (self.components) {
                // dispose previous items
                disposeComponents();
            }
            self.inputComponentOrganicOrder = [];
            self.idCounter = -1;
            self.components = [];
            self.schema.forEach(function (header, index) {
                var t = PivotForm.prototype.components.string,
                    id = getNewId(),
                    component;
                if (PivotForm.prototype.components[header.type]) {
                    t = PivotForm.prototype.components[header.type];
                }
                // instantiate component
                component = t.call(intf, header, index, intf);
                // add events, properties, id, name, index
                util.addEvents(component, header.componentEvents);
                component.header = header;
                component.name = component.name || (header.name === undefined ? '__' + id : header.name);
                component.id = id;
                component.index = index;
                // add component to the stack
                self.components.push(component);
                // listen to the component's change event
                function change() {
                    var cData = {};
                    // set the new value of the element to data
                    cData[header.name] = component.input.value;
                    //intf.data = cData;
                    // update any other components of the same header.name with this value
                    getComponentsByPropertyValue('name', header.name).forEach(function (comp) {
                        if (comp !== component) {
                            comp.value = cData;
                        }
                    });
                    // dispatch the change event to our listeners
                    intf.dispatchEvent(new Event('change'));
                }
                component.addEventListener('change', change);
                // if the component does not define the property value, define it
                if (!component.hasOwnProperty('value')) {
                    Object.defineProperty(component, 'value', {
                        get: function () {
                            // if the component defines an input, grab the value from the input's getter
                            if (component.input) {
                                return component.input.value;
                            }
                            // or just return the value set in data
                            return self.data[header.name];
                        },
                        // use the async value setter for set
                        set: function (value) {
                            // value is the entire data object
                            // if the value set has no property for this component's header name
                            // exit now
                            if (!value || !value.hasOwnProperty(header.name)) { return; }
                            // update value to be the actual value instead of the entire data object
                            value = value[header.name];
                            function setValue(v) {
                                // set the value of the form
                                intf.data[header.name] = v;
                                // if the component has a value to update, update it now
                                if (component.input) {
                                    component.input.value = v;
                                }
                                // if this component has not been initialized (had any initial value set)
                                // mark it as initialized to allow external data to begin flowing into it
                                if (!component.initialized) {
                                    component.initialized = true;
                                    if (component.classList) {
                                        component.classList.remove('loading');
                                    }
                                    if (component.dispatchEvent) {
                                        // tell dom initializer this component has been initialized
                                        // delay by one animation frame to allow for
                                        // other async setters in the schema to setup their functions
                                        requestAnimationFrame(function () {
                                            component.dispatchEvent(new Event('initialized'));
                                        });
                                    }
                                }
                            }
                            var r,
                                syncReturn,
                                // if the value passed in was a function, this is the callback for that function
                                callback = function (asyncValue) {
                                    // check if the function returned a value earlier, if so, it's probably a mistake
                                    if (syncReturn) {
                                        throw new Error('Async return detected from value already set via sync function.');
                                    }
                                    // set the value asynchronously via setter
                                    setValue(asyncValue);
                                };
                            // if a function was passed in, it could be async or sync
                            if (typeof value === 'function') {
                                if (component.classList) {
                                    component.classList.add('loading');
                                }
                                r = value.apply(component, [callback]);
                                // sync functions return a value now, async must return undefined (no return or return;)
                                syncReturn = r !== undefined;
                                if (syncReturn) {
                                    // if it was a sync value, run setter now
                                    setValue(r);
                                    return;
                                }
                                component.async = true;
                                return;
                            }
                            // if this was was not a function, run setter now
                            setValue(value);
                        }
                    });
                }
                util.setProperties(component.style, header.componentStyle);
            });
            createDom();
        }
        function dispatchResize() {
            intf.dispatchEvent(new Event('resize'));
        }
        function setMode() {
            if (intf.mode === 'dialog') {
                self.dialogOptions.title = intf.title;
                if (!self.dialog) {
                    self.dialog = dialog(self.dialogOptions);
                }
                // TODO, prevent double resize binding
                self.dialog.addEventListener('resize', dispatchResize);
                if (self.userStyleSheet) {
                    self.shadowRoot.appendChild(self.userStyleSheet);
                }
                self.dialog.content.appendChild(self.form);
                self.shadowRoot.appendChild(self.dialog);
                self.dialog.attached = true;
                if (self.shadowRoot.contains(self.fieldset)) { self.shadowRoot.removeChild(self.fieldset); }
            } else if (intf.mode === 'fieldset') {
                if (!self.fieldset) {
                    self.fieldset = util.createElement('fieldset', self.shadowRoot, {className: 'fieldset'});
                }
                self.fieldset.appendChild(self.legend);
                self.fieldset.appendChild(self.form);
                if (self.shadowRoot.contains(self.dialog)) { self.shadowRoot.removeChild(self.dialog); }
            } else {
                if (self.userStyleSheet) {
                    self.shadowRoot.appendChild(self.userStyleSheet);
                }
                self.shadowRoot.appendChild(self.form);
                if (self.shadowRoot.contains(self.fieldset)) { self.shadowRoot.removeChild(self.fieldset); }
                if (self.shadowRoot.contains(self.dialog)) { self.shadowRoot.removeChild(self.dialog); }
            }
        }
        function init() {
            if (intf.initialized) { return; }
            self.shadowRoot = intf.attachShadow ? intf.attachShadow({mode: 'open'}) : intf;
            if (!self.cssAttached) {
                util.createElement('link', self.shadowRoot, {
                    rel: 'stylesheet',
                    href: /^file:\/\/\//.test(window.location) ? '../lib/index.css' : 'data:text/css;base64,cssurl-target'
                });
                self.cssAttached = true;
            }
            setMode();
            intf.form = self.form;
            intf.initialized = true;
            loadState();
            if (self.dialog) {
                setTimeout(function () { self.dialog.centerHorizontally(); }, 1);
            }
        }
        function dataSetter(value) {
            if (!self.initialized) {
                Object.assign(self.preInitData, value);
                intf.childForms.forEach(function (form) {
                    form.data = value;
                });
                return;
            }
            initData(value);
        }
        function dataGetter() {
            var d = {};
            function componentDataGetter(component) {
                if (!component.header.name) { return; }
                d[component.header.name] = component.value;
            }
            self.components.forEach(componentDataGetter);
            intf.childForms.forEach(function (form) {
                form.components.forEach(componentDataGetter);
            });
            Object.keys(d).forEach(function (key) {
                var val = d[key];
                Object.defineProperty(d, key, {
                    set: function (value) {
                        var cData = {};
                        cData[key] = value;
                        if (!self.initialized) {
                            Object.assign(self.data, cData);
                            self.components.forEach(function (component) {
                                component.value = self.data;
                            });
                        }
                        intf.childForms.forEach(function (form) {
                            form.data = cData;
                        });
                    },
                    get: function () {
                        return val;
                    }
                });
            });
            return d;
        }
        function schemaSetter(value) {
            self.schema = value;
            initSchema();
        }
        function addChildForm(form) {
            form.id = getNewId();
            function childFormChange() {
                // tell the parent to dispatch an even to its listeners
                intf.dispatchEvent(new Event('change'));
            }
            form.name = self.name ? (self.name + '_pivot-form_' + self.childForms.length) : undefined;
            form.addEventListener('change', childFormChange);
            intf.addEventListener('stylesheetchanged', function () {
                form.stylesheet = intf.stylesheet;
            });
            if (intf.stylesheet) {
                form.stylesheet = intf.stylesheet;
            }
            self.initializingComponents[form.id] = form;
            form.addEventListener('initialized', function () {
                delete self.initializingComponents[form.id];
                doneInitializing();
            });
            self.childForms.push(form);
        }
        // public interface definitions
        intf.init = init;
        intf.dispose = dispose;
        intf.getComponentByName = getComponentByName;
        intf.getComponentById = getComponentById;
        intf.getComponentsByPropertyValue = getComponentsByPropertyValue;
        intf.addChildForm = addChildForm;
        Object.defineProperty(intf, 'childForms', {
            get: function () {
                var forms = [];
                function addForm(form) {
                    forms.push(form);
                    forms = forms.concat(form.childForms);
                }
                self.childForms.forEach(addForm);
                return forms;
            }
        });
        Object.defineProperty(intf, 'isContainer', {
            get: function () { return true; }
        });
        Object.defineProperty(intf, 'components', {
            get: function () {
                return self.components;
            }
        });
        Object.defineProperty(intf, 'data', {
            get: dataGetter,
            set: dataSetter
        });
        Object.defineProperty(intf, 'mode', {
            get: function () {
                return self.mode;
            },
            set: function (value) {
                var validModes = {
                    dialog: 'dialog',
                    block: 'block',
                    fieldset: 'fieldset'
                };
                self.mode = validModes[value] || 'block';
                if (!self.initialized) { return; }
                setMode();
            }
        });
        Object.defineProperty(intf, 'initializingComponents', {
            get: function () {
                return self.initializingComponents;
            }
        });
        Object.defineProperty(intf, 'title', {
            get: function () {
                return self.title;
            },
            set: function (value) {
                self.title = value;
                self.legend.innerHTML = self.title;
                if (self.dialog) {
                    self.dialog.title.innerHTML = value;
                }
            }
        });
        Object.defineProperty(intf, 'dialog', {
            get: function () {
                return self.dialog || { attached: false };
            }
        });
        Object.defineProperty(intf, 'schema', {
            get: function () {
                return self.schema;
            },
            set: schemaSetter
        });
        Object.defineProperty(intf, 'name', {
            get: function () {
                return self.name;
            },
            set: function (value) {
                self.name = value;
                loadState();
            }
        });
        Object.defineProperty(intf, 'stylesheet', {
            get: function () {
                if (!self.userStyleSheet) {
                    return undefined;
                }
                return self.userStyleSheet.originalValue;
            },
            set: function (value) {
                if (self.userStyleSheet && self.userStyleSheet.parentNode) {
                    self.userStyleSheet.parentNode.removeChild(self.userStyleSheet);
                    self.userStyleSheet = undefined;
                }
                if (!value) { return; }
                self.userStyleSheet = util.createElement('link');
                self.userStyleSheet.rel = 'stylesheet';
                self.userStyleSheet.type = 'text/css';
                self.userStyleSheet.href = value;
                self.userStyleSheet.originalValue = value;
                intf.dispatchEvent(new Event('stylesheetchanged'));
                if (self.shadowRoot) {
                    self.shadowRoot.appendChild(self.userStyleSheet);
                }
            }
        });
        // events
        window.addEventListener('resize', dispatchResize);
        return intf;
    }
    function connectedCallback(el) {
        (el || this).init();
        return;
    }
    PivotForm.prototype = Object.create(window.HTMLElement.prototype);
    if (window.customElements) {
        PivotForm.prototype.disconnectedCallback = function () {
            return;
        };
        PivotForm.prototype.getObservableAttributes = function () {
            return attributeList;
        };
        PivotForm.prototype.connectedCallback = connectedCallback;
        PivotForm.prototype.adoptedCallback = function () {
            return;
        };
        PivotForm.prototype.attributeChangedCallback = function (attrName, oldVal, newVal) {
            this[attrName] = newVal;
            return;
        };
        window.customElements.define('pivot-form', PivotForm);
    }
    function pivotFormFactory(args) {
        args = args || {};
        requestAnimationFrame(function () {
            window.dispatchEvent(new Event('resize'));
        });
        if (window.customElements) {
            return new PivotForm(args);
        }
        var f = new PivotForm(args);
        if (args.parentNode) {
            args.parentNode.appendChild(f.form);
        }
        attributeList.forEach(function (attrKey) {
            if (args[attrKey]) {
                f[attrKey] = args[attrKey];
            }
        });
        if (!f.form) {
            connectedCallback(f);
        }
        return f;
    }
    PivotForm.prototype.components = defaultComponents;
    window.pivotForm = pivotFormFactory;
    return pivotFormFactory;
});
